@@ -1,4 +1,4 @@
-/* $Id: rbldnsd_packet.c,v 1.86 2004/10/24 17:29:48 mjt Exp $
+/* $Id: rbldnsd_packet.c,v 1.96 2005/04/28 21:05:20 mjt Exp $
  * DNS packet handling routines for rbldnsd
  */
 
@@ -13,6 +13,18 @@
 #include "rbldnsd.h"
 #include "rbldnsd_hooks.h"
 
+#ifndef NOIPv6
+# ifndef NI_MAXHOST
+#  define IPSIZE 1025
+# else
+#  define IPSIZE NI_MAXHOST
+# endif
+#else
+# define IPSIZE 16
+#endif
+
+#define MAX_GLUE (MAX_NS*2)
+
 static int addrr_soa(struct dnspacket *pkt, const struct zone *zone, int auth);
 static int addrr_ns(struct dnspacket *pkt, const struct zone *zone, int auth);
 static int version_req(struct dnspacket *pkt, const struct dnsquery *qry);
@@ -54,30 +66,33 @@ static int version_req(struct dnspacket *pkt, const struct dnsquery *qry);
  * next two bytes are query class (IN, HESIOD etc)
  */
 
-/* since all counts are <255 due to size constraints, define
- * aliases for 2nd bytes */
-#define p_qdcnt p_qdcnt2
-#define p_ancnt p_ancnt2
-#define p_nscnt p_nscnt2
-#define p_arcnt p_arcnt2
-
-/* parsequery: parse query packet
- * initializes q_dn, q_dnlen, q_dnlab, q_type, q_class
- * returns pointer after qDN (where answer section will begin)
- * or NULL on failure. */
-
-static unsigned char *
-parsequery(register const unsigned char *q, unsigned qlen,
+static int
+parsequery(struct dnspacket *pkt, unsigned qlen,
            struct dnsquery *qry) {
 
   /* parsing incoming query.  Untrusted data read directly from the network.
-   * q is a buffer ptr - data that was read (DNS_MAXPACKET max).
+   * pkt->p_buf is a buffer - data that was read (DNS_EDNS0_MAXPACKET max).
    * qlen is number of bytes actually read (packet length)
    * first p_hdrsize bytes is header, next is query DN,
    * next are QTYPE and QCLASS (2x2 bytes).
-   * rest of data if any is ignored.
+   * If NSCNT==0 && ARCNT==1, and an OPT record comes after the query,
+   * EDNS0 packet size gets extracted from the OPT record.
+   * Rest of data is ignored.
+   * Returns true on success, 0 on failure.
+   * Upon successeful return, pkt->p_sans = pkt->p_cur points to the end of
+   * the query section (where our answers will be placed), and
+   * pkt->p_endp is initialized to point to the real end of answers.
+   * Real end of answers is:
+   * for non-EDNS0-aware clients it's pkt->p_buf+DNS_MAXPACKET, and
+   * if a vaild EDNS0 UDPsize is given, it will be pkt->p_buf+UDPsize-11,
+   * with the 11 bytes needed for a minimal OPT record.
+   * In replypacket() we check whenever all our answers fits in standard
+   * UDP buffer size (DNS_MAXPACKET), and if not (which means we're replying
+   * to EDNS0-aware client due to the above rules), we just add proper OPT
+   * record at the end.
    */
 
+  register unsigned const char *q = pkt->p_buf;
   register unsigned const char *x, *e;
   register unsigned char *d;
   unsigned qlab;			/* number of labels in qDN */
@@ -86,14 +101,14 @@ parsequery(register const unsigned char *q, unsigned qlen,
   /* qlen isn't needed anymore, it'll be used as length of qDN below */
 
   if (q + p_hdrsize > x)	/* short packet (header isn't here) */
-    return NULL;
+    return 0;
   else if (q + p_hdrsize + DNS_MAXDN <= x)
     x = q + p_hdrsize + DNS_MAXDN - 1; /* constrain query DN to DNS_MAXDN */
 
   if (q[p_f1] & pf1_qr)			/* response packet?! */
-     return 0;
+    return 0;
   if (q[p_qdcnt1] || q[p_qdcnt2] != 1)	/* qdcount should be == 1 */
-    return NULL;
+    return 0;
 
   /* parse and lowercase query DN, count and init labels */
   qlab = 0;			/* number of labels so far */
@@ -104,7 +119,7 @@ parsequery(register const unsigned char *q, unsigned qlen,
     e = q + *q + 1;		/* end of this label */
     if (*q > DNS_MAXLABEL	/* too long label? */
         || e > x)		/* or it ends past packet? */
-      return NULL;
+      return 0;
     /* lowercase it */
     ++q;			/* length */
     do *d++ = dns_dnlc(*q);	/* lowercase each char */
@@ -119,7 +134,29 @@ parsequery(register const unsigned char *q, unsigned qlen,
   qry->q_type = ((unsigned)(q[0]) << 8) | q[1];
   qry->q_class = ((unsigned)(q[2]) << 8) | q[3];
 
-  return (unsigned char*)q + 4;	/* answers will start here */
+  q += 4;
+  pkt->p_sans = (unsigned char *)q; /* answers will start here */
+  pkt->p_cur = (unsigned char *)q;  /* and current answer pointer is here */
+  d = pkt->p_buf;
+  if (q < x &&
+      d[p_nscnt1] == 0 && d[p_nscnt2] == 0 &&
+      d[p_arcnt1] == 0 && d[p_arcnt2] == 1 &&
+      q[0] == 0 /* empty DN */ &&
+      q[1] == (DNS_T_OPT>>8) && q[2] == (DNS_T_OPT&255)) {
+    qlen = (((unsigned)q[3]) << 8) | q[4];
+    /* 11 bytes are needed to encode minimal EDNS0 OPT record */
+    if (qlen < DNS_MAXPACKET + 11)
+      qlen = DNS_MAXPACKET;
+    else if (qlen > sizeof(pkt->p_buf) - 11)
+      qlen = sizeof(pkt->p_buf) - 11;
+    else
+      qlen -= 11;
+    pkt->p_endp = d + qlen;
+  }
+  else
+    pkt->p_endp = d + DNS_MAXPACKET;
+
+  return 1;
 }
 
 #ifdef RECOGNIZE_IP4IN6
@@ -230,12 +267,24 @@ int replypacket(struct dnspacket *pkt, unsigned qlen, struct zone *zone) {
 
   struct dnsquery qry;			/* query structure */
   struct dnsqinfo qi;			/* query info structure */
-  unsigned char *const h = pkt->p_buf;	/* packet's header */
+  unsigned char *h = pkt->p_buf;	/* packet's header */
   const struct dslist *dsl;
   int found;
   extern int lazy; /*XXX hack*/
 
-  if (!(pkt->p_cur = pkt->p_sans = parsequery(h, qlen, &qry))) {
+  pkt->p_substrr = 0;
+  /* check global ACL */
+  if (g_dsacl && g_dsacl->ds_stamp) {
+    found = ds_acl_query(g_dsacl, pkt);
+    if (found & NSQUERY_IGNORE) {
+      do_stats(gstats.q_err += 1; gstats.b_in += qlen);
+      return 0;
+    }
+  }
+  else
+    found = 0;
+
+  if (!parsequery(pkt, qlen, &qry)) {
     do_stats(gstats.q_err += 1; gstats.b_in += qlen);
     return 0;
   }
@@ -285,6 +334,7 @@ int replypacket(struct dnspacket *pkt, unsigned qlen, struct zone *zone) {
       refuse(DNS_R_NOTIMPL);
     qi.qi_tflag = NSQUERY_OTHER;
   }
+  qi.qi_tflag |= found;
   h[p_f2] = DNS_R_NOERROR;
 
   /* find matching zone */
@@ -298,9 +348,20 @@ int replypacket(struct dnspacket *pkt, unsigned qlen, struct zone *zone) {
 #define refuse(code)  _refuse(code, err_z)
   do_stats(zone->z_stats.b_in += qlen);
 
+  if (zone->z_dsacl && zone->z_dsacl->ds_stamp) {
+    qi.qi_tflag |= ds_acl_query(zone->z_dsacl, pkt);
+    if (qi.qi_tflag & NSQUERY_IGNORE) {
+      do_stats(gstats.q_err += 1);
+      return 0;
+    }
+  }
+
   if (!zone->z_stamp)	/* do not answer if not loaded */
     refuse(DNS_R_SERVFAIL);
 
+  if (qi.qi_tflag & NSQUERY_REFUSE)
+    refuse(DNS_R_REFUSED);
+
 #ifdef do_hook_query_access
   if ((found = hook_query_access(zone, NULL, &qi))) {
     if (found < 0) return 0;
@@ -310,8 +371,9 @@ int replypacket(struct dnspacket *pkt, unsigned qlen, struct zone *zone) {
 
   if (qi.qi_dnlab == 0) {	/* query to base zone: SOA and NS */
 
-    found = 1;
+    found = NSQUERY_FOUND;
 
+    /* NS and SOA with auth=0 will only touch answer section */
     if ((qi.qi_tflag & NSQUERY_SOA) && !addrr_soa(pkt, zone, 0))
       found = 0;
     else
@@ -319,7 +381,7 @@ int replypacket(struct dnspacket *pkt, unsigned qlen, struct zone *zone) {
       found = 0;
     if (!found) {
       pkt->p_cur = pkt->p_sans;
-      h[p_ancnt] = h[p_nscnt] = 0;
+      h[p_ancnt2] = h[p_nscnt2] = 0;
       refuse(DNS_R_REFUSED);
     }
 
@@ -329,10 +391,25 @@ int replypacket(struct dnspacket *pkt, unsigned qlen, struct zone *zone) {
 
   /* search the datasets */
   for(dsl = zone->z_dsl; dsl; dsl = dsl->dsl_next)
-    if (dsl->dsl_queryfn(dsl->dsl_ds, &qi, pkt))
-      found = 1;	/* positive answer */
+    found |= dsl->dsl_queryfn(dsl->dsl_ds, &qi, pkt);
+
+  if (found & NSQUERY_ADDPEER) {
+#ifdef NOIPv6
+    addrr_a_txt(pkt, qi.qi_tflag, pkt->p_substrr,
+                inet_ntoa(((struct sockaddr_in*)pkt->p_peer)->sin_addr),
+                pkt->p_substds);
+#else
+    char subst[IPSIZE];
+    if (getnameinfo(pkt->p_peer, pkt->p_peerlen,
+                    subst, NI_MAXHOST, NULL, 0, NI_NUMERICHOST) != 0)
+      subst[0] = '\0';
+    addrr_a_txt(pkt, qi.qi_tflag, pkt->p_substrr, subst, pkt->p_substds);
+#endif
+  }
 
   /* now complete the reply: add AUTH etc sections */
+  /* addrr_ns(auth=1) should be called last as it fills in
+   * both AUTH and ADDITIONAL sections */
   if (!found) {			/* negative result */
     addrr_soa(pkt, zone, 1);	/* add SOA if any to AUTHORITY */
     h[p_f2] = DNS_R_NXDOMAIN;
@@ -342,11 +419,11 @@ int replypacket(struct dnspacket *pkt, unsigned qlen, struct zone *zone) {
 #endif
   }
   else {
-    if (!h[p_ancnt]) {	/* positive reply, no answers */
+    if (!h[p_ancnt2]) {	/* positive reply, no answers */
       addrr_soa(pkt, zone, 1);	/* add SOA if any to AUTHORITY */
     }
     else if (zone->z_nns &&
-             (!(qi.qi_tflag & NSQUERY_NS) || qi.qi_dnlab) &&
+             /* (!(qi.qi_tflag & NSQUERY_NS) || qi.qi_dnlab) && */
              !lazy)
       addrr_ns(pkt, zone, 1); /* add nameserver records to positive reply */
     do_stats(zone->z_stats.q_ok += 1);
@@ -354,6 +431,20 @@ int replypacket(struct dnspacket *pkt, unsigned qlen, struct zone *zone) {
     hook_query_result(zone, NULL, &qi, 1);
 #endif
   }
+  if (rlen() > DNS_MAXPACKET) {	/* add OPT record for long replies */
+    /* as per parsequery(), we always have 11 bytes for minimal OPT record at
+     * the end of our reply packet, OR rlen() does not exceed DNS_MAXPACKET */
+    h[p_arcnt2] += 1;		/* arcnt is limited to 254 records */
+    h = pkt->p_cur;
+    *h++ = 0;			/* empty (root) DN */
+    PACK16S(h, DNS_T_OPT);
+    PACK16S(h, DNS_EDNS0_MAXPACKET);
+    *h++ = 0; *h++ = 0;		/* RCODE and version */
+    *h++ = 0; *h++ = 0;		/* rest of the TTL field */
+    *h++ = 0; *h++ = 0;		/* RDLEN */
+    pkt->p_cur = h;
+    h = pkt->p_buf;		/* restore for rlen() to work */
+  }
   do_stats(zone->z_stats.b_out += rlen());
   return rlen();
 
@@ -366,7 +457,7 @@ err_z:
   return rlen();
 }
 
-#define fit(p, c, bytes) ((c) + (bytes) <= (p)->p_buf + DNS_MAXPACKET)
+#define fit(pkt, c, bytes) ((c) + (bytes) <= (pkt)->p_endp)
 
 
 /* DN compression pointers/structures */
@@ -515,14 +606,21 @@ struct zonesoa {	/* cached SOA RR */
 };
 
 struct zonens {		/* cached NS RRs */
-  unsigned size;		/* total size of all NS RRs */
-  struct dnjump jump[MAX_NS*2];	/* jumps: for qDNs and for NSes */
-  struct dnjump *jend;		/* last jump */
+  unsigned nssize;			/* size of all NS RRs */
+  unsigned tsize;			/* size of NS+glue recs */
+  struct dnjump jump[MAX_NS*2+MAX_GLUE];/* jumps: for qDNs and for NSes */
+  struct dnjump *nsjend;		/* last NS jump */
+  struct dnjump *tjend;			/* last glue jump */
   unsigned char data[CACHEBUF_SIZE];
 };
 
 void init_zones_caches(struct zone *zonelist) {
   while(zonelist) {
+    if (!zonelist->z_dsl) {
+      char name[DNS_MAXDOMAIN];
+      dns_dntop(zonelist->z_dn, name, sizeof(name));
+      error(0, "missing data for zone `%s'", name);
+    }
     zonelist->z_zsoa = tmalloc(struct zonesoa);
     /* for NS RRs, we allocate MAX_NS caches:
      * each stores one variant of NS rotation */
@@ -584,35 +682,88 @@ static int addrr_soa(struct dnspacket *pkt, const struct zone *zone, int auth) {
   }
   /* for AUTHORITY section for NXDOMAIN etc replies, use minttl as TTL */
   if (auth) memcpy(c + zsoa->ttloff, zsoa->minttl, 4);
-  pkt->p_buf[auth ? p_nscnt : p_ancnt]++;
+  pkt->p_buf[auth ? p_nscnt2 : p_ancnt2]++;
   return 1;
 }
 
-int update_zone_ns(struct zone *zone, const struct dsns *dsns, unsigned ttl) {
+static unsigned char *
+find_glue(struct zone *zone, const unsigned char *nsdn,
+          struct dnspacket *pkt, const struct zone *zonelist) {
+  struct dnsqinfo qi;
+  unsigned lab;
+  unsigned char dnbuf[DNS_MAXDN], *dp;
+  unsigned char *dnlptr[DNS_MAXLABELS];
+  const struct dslist *dsl;
+  const struct zone *qzone;
+
+  /* lowercase the nsdn and find label pointers */
+  lab = 0; dp = dnbuf;
+  while((*dp = *nsdn)) {
+    const unsigned char *e = nsdn + *nsdn + 1;
+    dnlptr[lab++] = dp++;
+    while(++nsdn < e)
+      *dp++ = dns_dnlc(*nsdn);
+  }
+
+  qzone = findqzone(zonelist, dp - dnbuf + 1, lab, dnlptr, &qi);
+  if (!qzone)
+    return NULL;
+
+  /* pefrorm fake query */
+  qi.qi_tflag = NSQUERY_A/*|NSQUERY_AAAA*/;
+  dp = pkt->p_cur;
+  for(dsl = qzone->z_dsl; dsl; dsl = dsl->dsl_next)
+    dsl->dsl_queryfn(dsl->dsl_ds, &qi, pkt);
+
+  if (dp == pkt->p_cur) {
+    char name[DNS_MAXDOMAIN];
+    dns_dntop(qi.qi_dn, name, sizeof(name));
+    zlog(LOG_WARNING, zone, "no glue record(s) for %.60s NS found", name);
+    return NULL;
+  }
+
+  return dp;
+}
+
+int update_zone_ns(struct zone *zone, const struct dsns *dsns, unsigned ttl,
+                   const struct zone *zonelist) {
   struct zonens *zns;
   unsigned char *cpos, *sizep;
   struct dncompr compr;
   unsigned size, i, ns, nns;
   const unsigned char *nsdna[MAX_NS];
   const unsigned char *dn;
+  unsigned char *nsrrs[MAX_NS], *nsrre[MAX_NS];
+  unsigned nglue;
+  struct dnspacket pkt;
 
-  nns = 0;
-  while(dsns) {
+  memset(&pkt, 0, sizeof(pkt));
+  pkt.p_sans = pkt.p_cur = pkt.p_buf + p_hdrsize;
+  pkt.p_endp = pkt.p_buf + CACHEBUF_SIZE + p_hdrsize;
+
+  for(nns = 0; dsns; dsns = dsns->dsns_next) {
     i = 0;
     while(i < nns && !dns_dnequ(nsdna[i], dsns->dsns_dn))
       ++i;
-    if (i >= nns) {
-      if (nns < MAX_NS)
-        nsdna[nns++] = dsns->dsns_dn;
-      else {
-        zlog(LOG_WARNING, zone,
-             "too many NS records specified, only first %d will be used",
-             MAX_NS);
-        break;
-      }
+    if (i < nns)
+      continue;
+    if (nns >= MAX_NS) {
+      zlog(LOG_WARNING, zone,
+           "too many NS records specified, only first %d will be used", MAX_NS);
+      break;
     }
-    dsns = dsns->dsns_next;
+    nsdna[nns] = dsns->dsns_dn;
+    if ((nsrrs[nns] = find_glue(zone, dsns->dsns_dn, &pkt, zonelist)))
+      nsrre[nns] = pkt.p_cur;
+    ++nns;
   }
+  /* number of additional records must not exceed 254: (room for EDNS0 OPT) */
+  nglue = pkt.p_buf[p_ancnt2];
+  if (pkt.p_buf[p_ancnt1] || nglue > 254)	/* too many glue recs */
+    return 0;
+  /* check if we have enouth dnjump slots */
+  if (nns * 2 + nglue > sizeof(zns->jump)/sizeof(zns->jump[0]))
+    return 0;
 
   memcpy(zone->z_nsdna, nsdna, nns * sizeof(nsdna[0]));
   memset(nsdna + nns, 0, (MAX_NS - nns) * sizeof(nsdna[0]));
@@ -639,11 +790,24 @@ int update_zone_ns(struct zone *zone, const struct dsns *dsns, unsigned ttl) {
       size = cpos - sizep - 2;
       PACK16(sizep, size);
     }
+    dnc_finish(&compr, cpos, &zns->nssize, &zns->nsjend);
 
-    dnc_finish(&compr, cpos, &zns->size, &zns->jend);
+    if (nglue)
+      for(i = 0; i < nns; ++i)
+        for(dn = nsrrs[i]; dn && dn < nsrre[i]; ) {
+          /* pack the glue record. dnjump, type+class, ttl, size (= 4 or 16) */
+          dn += 2;
+          size = 10 + dn[2+2+4+1];
+          cpos = dnc_add(&compr, cpos, zone->z_nsdna[i]);
+          if (!cpos || cpos + size > compr.bend) return 0;
+          memcpy(cpos, dn, size);
+          dn += size; cpos += size;
+        }
+    dnc_finish(&compr, cpos, &zns->tsize, &zns->tjend);
 
     if (++ns >= nns) break;
 
+    /* rotate list of NSes */
     dn = nsdna[0];
     memmove(nsdna, nsdna + 1, (nns - 1) * sizeof(nsdna[0]));
     nsdna[nns - 1] = dn;
@@ -651,6 +815,7 @@ int update_zone_ns(struct zone *zone, const struct dsns *dsns, unsigned ttl) {
 
   }
   zone->z_nns = nns;
+  zone->z_nglue = nglue;
 
   return 1;
 }
@@ -660,9 +825,21 @@ static int addrr_ns(struct dnspacket *pkt, const struct zone *zone, int auth) {
   const struct zonens *zns = zone->z_zns + cns;
   if (!zone->z_nns)
     return 0;
-  if (!dnc_final(pkt, zns->data, zns->size, zns->jump, zns->jend))
+  /* if auth=1, we're adding last records (except maybe EDNS0 OPT),
+   * so it's ok to fill in both AUTH and ADDITIONAL sections. */
+  /* If we can't fit both NS and glue recs, try NS only, omitting glue.
+   * For auth=0, don't add glue records at all.  */
+  if (auth && dnc_final(pkt, zns->data, zns->tsize, zns->jump, zns->tjend)) {
+    pkt->p_buf[p_nscnt2] += zone->z_nns;
+    pkt->p_buf[p_arcnt2] += zone->z_nglue;
+  }
+  else if (!dnc_final(pkt, zns->data, zns->nssize, zns->jump, zns->nsjend))
     return 0;
-  pkt->p_buf[auth ? p_nscnt : p_ancnt] += zone->z_nns;
+  else
+    /* we can't overflow p_ancnt2 (255 max) because addrr_ns(auth=0)
+     * is called before all other answers will be collected,
+     * and MAX_NS (zone->z_nns) is definitely less than 255 */
+    pkt->p_buf[auth ? p_nscnt2 : p_ancnt2] += zone->z_nns;
   /* pick up next variation of NS ordering */
   ++cns;
   if (cns >= zone->z_nns)
@@ -737,7 +914,7 @@ void addrr_any(struct dnspacket *pkt, unsigned dtp,
   ttl = checkrr_present(pkt->p_sans, c, dtp, data, dsz, ttl);
   if (!ttl) return; /* if RR is already present, do nothing */
 
-  if (!fit(pkt, c, 12 + dsz)) {
+  if (!fit(pkt, c, 12 + dsz) || pkt->p_buf[p_ancnt2] == 255) {
     setnonauth(pkt->p_buf); /* non-auth answer as we can't fit the record */
     return;
   }
@@ -748,7 +925,7 @@ void addrr_any(struct dnspacket *pkt, unsigned dtp,
   PACK16S(c, dsz);
   memcpy(c, data, dsz);
   pkt->p_cur = c + dsz;
-  pkt->p_buf[p_ancnt] += 1; /* increment numanswers */
+  pkt->p_buf[p_ancnt2] += 1; /* increment numanswers */
 }
 
 void
@@ -787,37 +964,25 @@ static int version_req(struct dnspacket *pkt, const struct dnsquery *qry) {
   *c++ = --dsz;
   memcpy(c, show_version, dsz);
   pkt->p_cur = c + dsz;
-  pkt->p_buf[p_ancnt] += 1; /* increment numanswers */
-  return 1;		
+  pkt->p_buf[p_ancnt2] += 1; /* increment numanswers */
+  return 1;
 }
 
-void logreply(const struct dnspacket *pkt,
-              const struct sockaddr *peeraddr,
-              int UNUSED peeraddrlen,
-              FILE *flog, int flushlog) {
-#ifndef NOIPv6
-# ifndef NI_MAXHOST
-#  define IPSIZE 1025
-# else
-#  define IPSIZE NI_MAXHOST
-# endif
-#else
-# define IPSIZE 16
-#endif
+void logreply(const struct dnspacket *pkt, FILE *flog, int flushlog) {
   char cbuf[DNS_MAXDOMAIN + IPSIZE + 50];
   char *cp = cbuf;
   const unsigned char *const q = pkt->p_sans - 4;
 
   cp += sprintf(cp, "%lu ", (unsigned long)time(NULL));
 #ifndef NOIPv6
-  if (getnameinfo(peeraddr, peeraddrlen,
+  if (getnameinfo(pkt->p_peer, pkt->p_peerlen,
                   cp, NI_MAXHOST, NULL, 0,
                   NI_NUMERICHOST) == 0)
     cp += strlen(cp);
   else
     *cp++ = '?';
 #else
-  strcpy(cp, ip4atos(ntohl(((struct sockaddr_in*)peeraddr)->sin_addr.s_addr)));
+  strcpy(cp, inet_ntoa(((struct sockaddr_in*)pkt->p_peer)->sin_addr.s_addr));
   cp += strlen(cp);
 #endif
   *cp++ = ' ';
@@ -826,7 +991,7 @@ void logreply(const struct dnspacket *pkt,
       dns_typename(((unsigned)q[0]<<8)|q[1]),
       dns_classname(((unsigned)q[2]<<8)|q[3]),
       dns_rcodename(pkt->p_buf[p_f2] & pf2_rcode),
-      pkt->p_buf[p_ancnt], pkt->p_cur - pkt->p_buf);
+      pkt->p_buf[p_ancnt2], pkt->p_cur - pkt->p_buf);
   if (flushlog)
     write(fileno(flog), cbuf, cp - cbuf);
   else
